@@ -153,13 +153,14 @@ class BaseMarshal:
     def to_python(self, proto_type, value, *, absent: bool = None):
         # Internal protobuf has its own special type for lists of values.
         # Return a view around it that implements MutableSequence.
-        if isinstance(value, compat.repeated_composite_types):
+        value_type = type(value)  # Minor performance boost over isinstance
+        if value_type in compat.repeated_composite_types:
             return RepeatedComposite(value, marshal=self)
-        if isinstance(value, compat.repeated_scalar_types):
+        if value_type in compat.repeated_scalar_types:
             return Repeated(value, marshal=self)
 
         # Same thing for maps of messages.
-        if isinstance(value, compat.map_composite_types):
+        if value_type in compat.map_composite_types:
             return MapComposite(value, marshal=self)
 
         # Convert ordinary values.
@@ -182,7 +183,7 @@ class BaseMarshal:
 
             # Convert lists and tuples recursively.
             if isinstance(value, (list, tuple)):
-                return type(value)([self.to_proto(proto_type, i) for i in value],)
+                return type(value)(self.to_proto(proto_type, i) for i in value)
 
         # Convert dictionaries recursively when the proto type is a map.
         # This is slightly more complicated than converting a list or tuple
@@ -196,9 +197,8 @@ class BaseMarshal:
             proto_type.DESCRIPTOR.has_options
             and proto_type.DESCRIPTOR.GetOptions().map_entry
         ):
-            return {
-                k: self.to_proto(type(proto_type().value), v) for k, v in value.items()
-            }
+            recursive_type = type(proto_type().value)
+            return {k: self.to_proto(recursive_type, v) for k, v in value.items()}
 
         # Convert ordinary values.
         rule = self._rules.get(proto_type, self._noop)
@@ -209,7 +209,8 @@ class BaseMarshal:
             raise TypeError(
                 "Parameter must be instance of the same class; "
                 "expected {expected}, got {got}".format(
-                    expected=proto_type.__name__, got=pb_value.__class__.__name__,
+                    expected=proto_type.__name__,
+                    got=pb_value.__class__.__name__,
                 ),
             )
 
@@ -235,9 +236,11 @@ class Marshal(BaseMarshal):
                 marshals with the same ``name`` argument will provide the
                 same marshal each time.
         """
-        if name not in cls._instances:
-            cls._instances[name] = super().__new__(cls)
-        return cls._instances[name]
+        klass = cls._instances.get(name)
+        if klass is None:
+            klass = cls._instances[name] = super().__new__(cls)
+
+        return klass
 
     def __init__(self, *, name: str):
         """Instantiate a marshal.
