@@ -68,7 +68,9 @@ class MessageMeta(type):
             # Determine the name of the entry message.
             msg_name = "{pascal_key}Entry".format(
                 pascal_key=re.sub(
-                    r"_\w", lambda m: m.group()[1:].upper(), key,
+                    r"_\w",
+                    lambda m: m.group()[1:].upper(),
+                    key,
                 ).replace(key[0], key[0].upper(), 1),
             )
 
@@ -84,20 +86,26 @@ class MessageMeta(type):
                 {
                     "__module__": attrs.get("__module__", None),
                     "__qualname__": "{prefix}.{name}".format(
-                        prefix=attrs.get("__qualname__", name), name=msg_name,
+                        prefix=attrs.get("__qualname__", name),
+                        name=msg_name,
                     ),
                     "_pb_options": {"map_entry": True},
                 }
             )
             entry_attrs["key"] = Field(field.map_key_type, number=1)
             entry_attrs["value"] = Field(
-                field.proto_type, number=2, enum=field.enum, message=field.message,
+                field.proto_type,
+                number=2,
+                enum=field.enum,
+                message=field.message,
             )
             map_fields[msg_name] = MessageMeta(msg_name, (Message,), entry_attrs)
 
             # Create the repeated field for the entry message.
             map_fields[key] = RepeatedField(
-                ProtoType.MESSAGE, number=field.number, message=map_fields[msg_name],
+                ProtoType.MESSAGE,
+                number=field.number,
+                message=map_fields[msg_name],
             )
 
         # Add the new entries to the attrs
@@ -107,18 +115,17 @@ class MessageMeta(type):
         # Iterate over all the attributes and separate the fields into
         # their own sequence.
         fields = []
+        new_attrs = {}
         oneofs = collections.OrderedDict()
         proto_imports = set()
         index = 0
-        for key, field in copy.copy(attrs).items():
+        for key, field in attrs.items():
             # Sanity check: If this is not a field, do nothing.
             if not isinstance(field, Field):
+                # The field objects themselves should not be direct attributes.
+                new_attrs[key] = field
                 continue
 
-            # Remove the field from the attrs dictionary; the field objects
-            # themselves should not be direct attributes.
-            attrs.pop(key)
-
             # Add data that the field requires that we do not take in the
             # constructor because we can derive it from the metaclass.
             # (The goal is to make the declaration syntax as nice as possible.)
@@ -178,29 +185,19 @@ class MessageMeta(type):
         #
         # m = MyMessage()
         # MyMessage.field in m
-        mcls = type("AttrsMeta", (mcls,), opt_attrs)
+        if opt_attrs:
+            mcls = type("AttrsMeta", (mcls,), opt_attrs)
 
         # Determine the filename.
         # We determine an appropriate proto filename based on the
         # Python module.
-        filename = "{0}.proto".format(
-            attrs.get("__module__", name.lower()).replace(".", "/")
+        filename = _file_info._FileInfo.proto_file_name(
+            new_attrs.get("__module__", name.lower())
         )
 
         # Get or create the information about the file, including the
         # descriptor to which the new message descriptor shall be added.
-        file_info = _file_info._FileInfo.registry.setdefault(
-            filename,
-            _file_info._FileInfo(
-                descriptor=descriptor_pb2.FileDescriptorProto(
-                    name=filename, package=package, syntax="proto3",
-                ),
-                enums=collections.OrderedDict(),
-                messages=collections.OrderedDict(),
-                name=filename,
-                nested={},
-            ),
-        )
+        file_info = _file_info._FileInfo.maybe_add_descriptor(filename, package)
 
         # Ensure any imports that would be necessary are assigned to the file
         # descriptor proto being created.
@@ -209,7 +206,7 @@ class MessageMeta(type):
                 file_info.descriptor.dependency.append(proto_import)
 
         # Retrieve any message options.
-        opts = descriptor_pb2.MessageOptions(**attrs.pop("_pb_options", {}))
+        opts = descriptor_pb2.MessageOptions(**new_attrs.pop("_pb_options", {}))
 
         # Create the underlying proto descriptor.
         desc = descriptor_pb2.DescriptorProto(
@@ -223,9 +220,9 @@ class MessageMeta(type):
 
         # If any descriptors were nested under this one, they need to be
         # attached as nested types here.
-        for child_path in copy.copy(file_info.nested).keys():
-            if local_path == child_path[:-1]:
-                desc.nested_type.add().MergeFrom(file_info.nested.pop(child_path),)
+        child_paths = [p for p in file_info.nested.keys() if local_path == p[:-1]]
+        for child_path in child_paths:
+            desc.nested_type.add().MergeFrom(file_info.nested.pop(child_path))
 
         # Add the descriptor to the file if it is a top-level descriptor,
         # or to a "holding area" for nested messages otherwise.
@@ -235,7 +232,7 @@ class MessageMeta(type):
             file_info.nested[local_path] = desc
 
         # Create the MessageInfo instance to be attached to this message.
-        attrs["_meta"] = _MessageInfo(
+        new_attrs["_meta"] = _MessageInfo(
             fields=fields,
             full_name=full_name,
             marshal=marshal,
@@ -244,7 +241,7 @@ class MessageMeta(type):
         )
 
         # Run the superclass constructor.
-        cls = super().__new__(mcls, name, bases, attrs)
+        cls = super().__new__(mcls, name, bases, new_attrs)
 
         # The info class and fields need a reference to the class just created.
         cls._meta.parent = cls
@@ -286,7 +283,13 @@ class MessageMeta(type):
             if coerce:
                 obj = cls(obj)
             else:
-                raise TypeError("%r is not an instance of %s" % (obj, cls.__name__,))
+                raise TypeError(
+                    "%r is not an instance of %s"
+                    % (
+                        obj,
+                        cls.__name__,
+                    )
+                )
         return obj._pb
 
     def wrap(cls, pb):
@@ -296,7 +299,10 @@ class MessageMeta(type):
             pb: A protocol buffer object, such as would be returned by
                 :meth:`pb`.
         """
-        return cls(pb, __wrap_original=True)
+        # Optimized fast path.
+        instance = cls.__new__(cls)
+        instance.__dict__["_pb"] = pb
+        return instance
 
     def serialize(cls, instance) -> bytes:
         """Return the serialized proto.
@@ -320,23 +326,26 @@ class MessageMeta(type):
             ~.Message: An instance of the message class against which this
             method was called.
         """
-        # Usually we don't wrap the original proto and are force to make a copy
-        # to prevent modifying user data.
-        # In this case it's perfectly reasonable to wrap the proto becasue it's
-        # never user visible, and it gives a huge performance boost.
-        return cls(cls.pb().FromString(payload), __wrap_original=True)
+        return cls.wrap(cls.pb().FromString(payload))
 
-    def to_json(cls, instance) -> str:
+    def to_json(cls, instance, *, use_integers_for_enums=True) -> str:
         """Given a message instance, serialize it to json
 
         Args:
             instance: An instance of this message type, or something
                 compatible (accepted by the type's constructor).
+            use_integers_for_enums (Optional(bool)): An option that determines whether enum
+                values should be represented by strings (False) or integers (True).
+                Default is True.
 
         Returns:
             str: The json string representation of the protocol buffer.
         """
-        return MessageToJson(cls.pb(instance))
+        return MessageToJson(
+            cls.pb(instance),
+            use_integers_for_enums=use_integers_for_enums,
+            including_default_value_fields=True,
+        )
 
     def from_json(cls, payload) -> "Message":
         """Given a json string representing an instance,
@@ -371,44 +380,46 @@ class Message(metaclass=MessageMeta):
         #   * A dict
         #   * Nothing (keyword arguments only).
 
-        # Handle the first two cases: they both involve keeping
-        # a copy of the underlying protobuf descriptor instance.
-        if isinstance(mapping, self._meta.pb):
+        if mapping is None:
+            if not kwargs:
+                # Special fast path for empty construction.
+                self.__dict__["_pb"] = self._meta.pb()
+                return
+
+            mapping = kwargs
+        elif isinstance(mapping, self._meta.pb):
             # Make a copy of the mapping.
             # This is a constructor for a new object, so users will assume
             # that it will not have side effects on the arguments being
             # passed in.
             #
-            # The `__wrap_original` argument is private API to override
-            # this behavior, because `MessageRule` actually does want to
-            # wrap the original argument it was given. The `wrap` method
-            # on the metaclass is the public API for this behavior.
-            if not kwargs.pop("__wrap_original", False):
-                mapping = copy.copy(mapping)
-            self._pb = mapping
+            # The `wrap` method on the metaclass is the public API for taking
+            # ownership of the passed in protobuf objet.
+            mapping = copy.copy(mapping)
             if kwargs:
-                self._pb.MergeFrom(self._meta.pb(**kwargs))
+                mapping.MergeFrom(self._meta.pb(**kwargs))
+
+            self.__dict__["_pb"] = mapping
             return
-        if isinstance(mapping, type(self)):
-            # Performance hack to streamline construction from vanilla protos.
+        elif isinstance(mapping, type(self)):
+            # Just use the above logic on mapping's underlying pb.
             self.__init__(mapping=mapping._pb, **kwargs)
             return
-
-        # Handle the remaining case by converging the mapping and kwargs
-        # dictionaries (with kwargs winning), and saving a descriptor
-        # based on that.
-
-        if mapping is None:
-            mapping = {}
         elif not isinstance(mapping, collections.abc.Mapping):
             # Sanity check: Did we get something not a map? Error if so.
             raise TypeError(
                 "Invalid constructor input for %s: %r"
-                % (self.__class__.__name__, mapping,)
+                % (
+                    self.__class__.__name__,
+                    mapping,
+                )
             )
-        mapping.update(kwargs)
+        else:
+            # Can't have side effects on mapping.
+            mapping = copy.copy(mapping)
+            # kwargs entries take priority for duplicate keys.
+            mapping.update(kwargs)
 
-        # Avoid copying the mapping unnecessarily
         params = {}
         # Update the mapping to address any values that need to be
         # coerced.
@@ -420,7 +431,7 @@ class Message(metaclass=MessageMeta):
                 params[key] = pb_value
 
         # Create the internal protocol buffer.
-        self._pb = self._meta.pb(**params)
+        self.__dict__["_pb"] = self._meta.pb(**params)
 
     def __bool__(self):
         """Return True if any field is truthy, False otherwise."""
@@ -572,8 +583,8 @@ class _MessageInfo:
         self.package = package
         self.full_name = full_name
         self.options = options
-        self.fields = collections.OrderedDict([(i.name, i) for i in fields])
-        self.fields_by_number = collections.OrderedDict([(i.number, i) for i in fields])
+        self.fields = collections.OrderedDict((i.name, i) for i in fields)
+        self.fields_by_number = collections.OrderedDict((i.number, i) for i in fields)
         self.marshal = marshal
         self._pb = None
 
